/***********************************************************************NTNU IT*
* TestSpeed.c - Linux x86 / x86_64 cpu cycle/byte testing of a NIST API
* hash function Hash(hashbitlen, data, databitlen, hashval).
*------------------------------------------------------------------------------
* $Id: TestSpeed-lx.c 69 2009-09-14 06:42:16Z joern $
*******************************************************************************/

#define _GNU_SOURCE
#include <sched.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <limits.h>
#include <unistd.h>
#include <sys/utsname.h>

#if !defined(LLONG_MAX)
#define LLONG_MAX (~0ull>>1)
#endif

#include "rdtsc.h"
#include "SHA3api_ref.h"

static void do_header(char *line, FILE *fp, int argc, char **argv);
#define teeprint(s,fp) (fputs(s,stdout), fputs(s,fp))

#define ST_OUTFILE "SpeedTest.txt"
#define ST_FILE_LBUF 254
#define ST_MD_SIZE  64
#define ST_NUM_TESTRUNS 10000


static int bitlens[] = {  /* the supported hash output sizes */
	224, 256, 384, 512, 0 
};
static int mdruns[] = {   /* standard and SUPERCOP benchmark message lengths */
	1, 8, 10, 64, 100, 576, 1000, 1536, 4096, 10000, 100000, 0
};


int
main(int argc, char **argv)
{
	BitSequence *Msg, *MD;
	FILE *fp_out;
	char *line;
	unsigned mdsum; /* checksum to countermeasure compiler optimizations */
	int i, max_md;


	/* constrain the process to one core in to obtain meaningful timing data */
	{
		cpu_set_t cpumask;
		int cpu;

		CPU_ZERO(&cpumask);
		if ((cpu = sched_getcpu()) != -1) {
			CPU_SET(cpu, &cpumask);
			if (sched_setaffinity(0, sizeof(cpu_set_t), &cpumask) == -1) {
				perror("\tsched_setaffinity()");
				exit(1);
			}
		}
	}

	/* find max message length and allocate speed test memory */
	max_md = mdruns[0];
	for (i = 0, max_md = mdruns[0]; mdruns[i] > 0; i++)
		if (mdruns[i] > max_md)
			max_md = mdruns[i];
	Msg = malloc(max_md+1);
	MD = malloc(ST_MD_SIZE);
	line = malloc(ST_FILE_LBUF+2);
	if (!Msg || !MD || !line) {
		fputs("No memory\n\n", stderr);
		exit(1);
	}

	/* open speed test output file */
	unlink(ST_OUTFILE);
	if ((fp_out = fopen(ST_OUTFILE, "w")) == NULL) {
		fputs("Couldn't open " ST_OUTFILE " for write\n", stderr);
		exit(1);
	}
  
	/* output header to terminal and file */
	do_header(line, fp_out, argc, argv);

	/* initialize message */
	for (i = 0; i < max_md; i++)
		Msg[i] = (BitSequence)i;

	
	/* output size loop */
	for (mdsum = 0, i = 0; bitlens[i] > 0; i++) {
		int j;

		/* message length loop */
		for(j = 0; mdruns[j] > 0; j++) {
			unsigned long long c, c_min;
			int k, n;

			/* run the test NUM_TESTRUNS times and find minimum cycle count */
			n = mdruns[j];
			c_min = LLONG_MAX;
			for(k = 0; k < ST_NUM_TESTRUNS; k++ ) {
				c = read_tsc();
				Hash(bitlens[i], Msg, n*8, MD);
				c = read_tsc() - c;
				if (c < c_min)
					c_min = c;
			}
			mdsum ^= *(unsigned *)MD;
			sprintf(line, "%6d  %8d  %8.3lf\n", bitlens[i], n,
				(c_min+1)/(double)n);
			teeprint(line, fp_out);
		}
	}

	/* print checksum and close output */
	sprintf(line, "\n\nmdsum 0x%x\n\n", mdsum);
	teeprint(line, fp_out);
	fclose(fp_out);

	return 0;
}


#define ARGCMP(s, arg) (strncmp(s, arg, strlen(s)))

static void
do_header(char *line, FILE *fp, int argc, char **argv)
{
	/* print header text */
	teeprint("\nBlueMidnightWish performance in cycles/byte with various "
		"message byte lengths\n", fp);
	sprintf(line, "               [ minimum of %d, results in units "
		"per byte ]\n\n", ST_NUM_TESTRUNS);
	teeprint(line, fp);

	/* print test date */
	{
		time_t t;
		struct tm *tp;

		time(&t);
		tp = localtime(&t);
		if (tp)
			sprintf(line, "date:     %4d-%02d-%02d %02d:%02d:%02d\n",
				tp->tm_year+1900, tp->tm_mon+1, tp->tm_mday, tp->tm_hour,
				tp->tm_min,	tp->tm_sec);
		else
			strcpy(line, "date:     <unknown>\n");

		teeprint(line, fp);
	}

	/* print cmd line arguments */
	{
		char *abi, *arch;

		abi = arch = "<unknown>";
		for (++argv; *argv; argv++) {
			if (ARGCMP("abi=", *argv) == 0)
				abi = *argv + strlen("abi=");
			else if (ARGCMP("arch=", *argv) == 0)
				arch = *argv + strlen("arch=");
		}
		sprintf(line, "arch:     %s\nabi:      %s\n", arch, abi);
		teeprint(line, fp);
	}

	/* print name and kernel */
	{
		struct utsname uts;

		if (uname(&uts) == 0)
#if defined(_AIX)
			sprintf(line, "hostname: %s\nmachine:  %s\nosrel:    %s.%s\n",
				uts.nodename, uts.machine, uts.version, uts.release);
#else
			sprintf(line, "hostname: %s\nmachine:  %s\nosrel:    %s\n",
				uts.nodename, uts.machine, uts.release);
#endif /*_AIX*/
		else
			strcpy(line, "hostname: <unknown>\nmachine:  <unknown>\n"
				"osrel:    <unknown>\n");
		teeprint(line, fp);
	}

	/* find and print compiler version */
	{
#if defined(__SUNPRO_C)
		int major, minor, micro;
#endif /*__SUNPRO_C*/

#if defined(__VERSION__) && !defined(__ICC)
		sprintf(line, "compiler: gcc %s\n", __VERSION__);
#elif defined(__xlc__)
		sprintf(line, "compiler: xlc %s\n", __xlc__);
#elif defined(__ICC)
		sprintf(line, "compiler: icc %d.%d.%d\n", __ICC/100, __ICC%100,
			__INTEL_COMPILER_BUILD_DATE);
#elif defined(__PGIC__)
		sprintf(line, "pgcc %d.%d.%d\n", __PGIC__, __PGIC_MINOR__,
			__PGIC_PATCHLEVEL__);
#elif defined(__SUNPRO_C)
		micro = __SUNPRO_C & 0xf;
		minor = (__SUNPRO_C >> 4) & 0xf;
		major = (__SUNPRO_C >> 8) & 0xf;

		if (micro)
			sprintf(line, "compiler: sun_c %d.%d.%d", major, minor, micro);
		else
			sprintf(line, "compiler: sun_c %d.%d", major, minor);
#else
		strcpy(line, "compiler: <unknown>");
#endif

		teeprint(line, fp);
	}

	/* finally print bytes/cycles header */
#if defined(_AIX)
	teeprint("\nMDSIZE  NUMBYTES    CYCLES     INSTR    INTOPS\n", fp);
#else
	teeprint("\nMDSIZE  NUMBYTES    CYCLES\n", fp);
#endif /*_AIX*/
}
